/*
 * Copyright (c) 2013-2023 NIBIO.
 *
 * This file is part of VIPSWeb 
 * (see https://gitlab.nibio.no/VIPS/VIPSWeb).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * Utilities and methods for using the IPM Decisions DSS and Weather APIs
 * 
 * Dependencies:
 * momentJS
 * turf 6
 * OpenLayers 4
 */

// TODO Refactor these to be set by the user of this library
//const ipmdDSSApiURL = "https://platform.ipmdecisions.net/api/dss/";
const ipmdDSSApiURL = "https://ipmdecisions.nibio.no/api/dss/";
const ipmdWeatherApiURL = "https://ipmdecisions.nibio.no/api/wx/";
//const ipmdWeatherApiURL = "http://localhost:18081/";


/**
 * 
 * @param {String} dssId The id of the DSS
 * @param {String} modelId The id of the DSS model
 * @returns {Json} metadata for the requested DSS model
 */
async function getModelMetadata(dssId,modelId) {
    const response = await fetch(ipmdDSSApiURL + "rest/dss/" + dssId);
    const dss = await response.json();
    for(let i=0; i< dss.models.length;i++)
    {
        let model = dss.models[i];
        if(model.id == modelId)
        {
            return model;
        }
    }
    return null;
}

/**
 * 
 * @param {String} dssId The id of the DSS
 * @param {String} modelId The id of the DSS model
 * @returns the UI form components of the Json schema for input data for the requested model. Hidden fields are excluded.
 */
async function getModelInputSchema(dssId,modelId) {
    const response = await fetch(ipmdDSSApiURL + "rest/model/" + dssId + "/" + modelId + "/input_schema/ui_form");
    return await response.json();
}

/**
 * 
 * @param {Json} weatherDatasource 
 * @returns The web service endpoint for the weather data source
 */
function getWeatherDatasourceEndpoint(weatherDatasource)
{
    return weatherDatasource.endpoint.replace("{WEATHER_API_URL}",ipmdWeatherApiURL);
}

/**
 * 
 * @param {Json} weatherDatasources The list of weather data sources
 * @param {String} weatherDatasourceId the id of the requested weather data source
 * @returns {Json} Metadata for the requested weather datasource
 */
function getWeatherDatasource(weatherDatasources, weatherDatasourceId)
{
    for(let i=0;i<weatherDatasources.length;i++)
    {
        if(weatherDatasources[i].id == weatherDatasourceId)
        {
            return weatherDatasources[i];
        }
    }
    return null;
}

/**
 * 
 * @returns {Json} list of all available weather datasources
 */
async function getWeatherDatasources()
{
    const response = await fetch(ipmdWeatherApiURL + "rest/weatherdatasource/");
    return await response.json();
}

/**
 * TODO Add more filtering than just authenticationType
 * @param {String} filterCriteria {"authenticationType": ["NONE", "CREDENTIALS", "BEARER_TOKEN"]}
 */
async function getFilteredWeatherDatasources(filterCriteria)
{
    let allWeatherDatasources = await getWeatherDatasources();
    let filteredWeatherDatasources = allWeatherDatasources;
    let authenticationType = filterCriteria["authenticationType"];
    if(authenticationType != undefined)
    {
        filteredWeatherDatasources = [];
        for(let i=0;i< allWeatherDatasources.length; i++)
        {
            if(allWeatherDatasources[i].authentication_type == authenticationType || (allWeatherDatasources[i].authentication_type == undefined && authenticationType == "NONE"))
            {
                filteredWeatherDatasources.push(allWeatherDatasources[i])
            }
        }
    }
    return filteredWeatherDatasources;
}

/**
 * 
 * @param {String} weatherDatasource 
 * @returns {Array} Weather stations listed as part of this weather data source. Example list item: {"id":"5", "name":"Merrapanna"}
 */
function getWeatherStationList(weatherDatasource){
    geoJson = JSON.parse(weatherDatasource.spatial.geoJSON);
    let stationList = [];
    for(let i=0;i<geoJson.features.length;i++)
    {
        let feature = geoJson.features[i];
        let station = {"id":feature.id, "name":feature.properties.name}
        stationList.push(station);
    }
    stationList.sort((a, b) => {
        return (a.name < b.name) ? -1 : (a.name > b.name) ?  1 : 0;
    });
    return stationList;
}

/**
 * List all weather stations in list in the provided HTML Form select list
 * @param {Array} stationList 
 * @param {HTMLElement} selectList 
 */
function renderWeatherStationSelectList(stationList, selectList){
    selectList.options.length = 0;
    selectList.add(new Option("-- Please select weather station --", -1));
    for(let i=0;i<stationList.length; i++)
    {
        selectList.add(new Option(stationList[i].name, stationList[i].id));
    }
}

/**
 * 
 * @param {String} endpoint URL to web service endpoint
 * @param {String} weatherStationId Id of the weather station
 * @param {Array} parameters List of weather parameters (IPM Decisions codes)
 * @param {Int} interval log interval for the weather data, provided in seconds. E.g. 3600 = hourly, 86400 = daily
 * @param {String} dateStart (YYYY-MM-DD)
 * @param {String} dateEnd (YYYY-MM-DD)
 * @returns weather data from a station based weather data source
 */
async function getStationWeatherData(endpoint, weatherStationId, parameters, interval, dateStart, dateEnd){
    const response = await fetch(endpoint 
        + "?timeStart="        + dateStart
        + "&timeEnd="          + dateEnd
        + "&interval="         + interval
        + "&weatherStationId=" + weatherStationId
        + "&parameters=" + parameters.join(",")
        );

        if(response.status == 200)
        {
            return await response.json();
        }
        else 
        {
            let errorMessage = "No weather data was returned. Message from the data source was: \n\n" + await response.text();
            alert(errorMessage);
            return null;
        }
}

/**
 * 
 * @param {String} endpoint URL to web service endpoint
 * @param {Float} longitude Decimal degrees (WGS84) for the location
 * @param {Float} latitude Decimal degrees (WGS84) for the location
 * @param {Array} parameters List of weather parameters (IPM Decisions codes)
 * @param {Int} interval log interval for the weather data, provided in seconds. E.g. 3600 = hourly, 86400 = daily
 * @param {String} dateStart (YYYY-MM-DD)
 * @param {String} dateEnd (YYYY-MM-DD)
 * @returns weather data from a location based weather data source
 */
async function getLocationWeatherData(endpoint, longitude, latitude, parameters, interval, dateStart, dateEnd){
    const response = await fetch(endpoint 
        + "?timeStart="        + dateStart
        + "&timeEnd="          + dateEnd
        + "&interval="         + interval
        + "&longitude="        + longitude
        + "&latitude="         + latitude
        + "&parameters=" + parameters.join(",")
        );

    if(response.status == 200)
    {
        return await response.json();
    }
    else 
    {
        let errorMessage = "No weather data was returned. Message from the data source was: \n\n" + await response.text();
        alert(errorMessage);
        return null;
    }
    
}

/**
 * Run and get results back from a DSS model
 * @param {String} endpoint The web service endpont for that DSS model
 * @param {Json} inputData as defined by the Json schema in the DSS model metadata property input_data
 * @returns {Json} model results in IPM Decisions format
 */
async function runModel(endpoint, inputData)
{
    const response = await fetch(endpoint, {
        method: "POST",
        mode: "cors",
        cache: "no-cache",
        headers: {
            "Content-Type": "application/json"
        },
        body: JSON.stringify(inputData)
    });
    return await response.json();
}

/**
 * Util model for getting an array of consecutive dates
 * @param {Integer} timestart UNIX epoch seconds
 * @param {Integer} interval log interval in seconds. E.g. 3600 = hourly, 86400 = daily
 * @param {Integer} length How many dates you want in the array
 * @returns 
 */
function getDateArray(timestart, interval, length)
{
    
    let dateArray = [];
    currentTime = moment(timestart);
    for(let i=0;i< length; i++)
    {
        dateArray.push(currentTime.valueOf());
        currentTime.add(interval,'seconds');
    }
    return dateArray;
}

/**
 * 
 * @param {Integer} weatherParameterId the parameter id
 * @returns {Json} parameter metadata (name, description, unit, aggregation type (MIN/MAX/AVG))
 */
function getWeatherParameter(weatherParameterId){
    for(let i=0;i<weatherParameterList.length;i++)
    {
        if(weatherParameterList[i].id == weatherParameterId)
        {
            return weatherParameterList[i];
        }
    }
    return null;
}

/**
 * 
 * @param {Json} weatherDatasource Metadata for the weather datasource
 * @param {String} weatherStationId the requested weather station
 * @returns {Array} coordinate (Decimal degrees/WGS84) of the weather station. 
 */
function getWeatherStationCoordinate(weatherDatasource, weatherStationId)
{
    geoJson = JSON.parse(weatherDatasource.spatial.geoJSON);
    let stationList = [];
    for(let i=0;i<geoJson.features.length;i++)
    {
        let feature = geoJson.features[i];
        if(feature.id == weatherStationId)
        {
            // TODO: Don't assume feature is a point
            return feature.geometry.coordinates;
        }
    }

    return null;
}

/**
 * Models may require weather parameters that the requested weather datasource
 * can't provide. The difference between a requested parameter and available parameter
 * may be insignificant. For instance, a model may require average temperature for an hour,
 * and the data source may provide instantaneous temperature for that hour. 
 * 
 * This methods figures out which parameters are considered fallback parameters for each other,
 * and returns the intersection of the requested parameters (including fallbacks) and the available 
 * parameters
 * 
 * @param {Array} requestedParameters the parameters that the model requires
 * @param {Array} availableParameters the parameters that the weather datasource provides
 * @returns 
 */
function getPragmaticWeatherParameterList(requestedParameters, availableParameters)
{
    return getParametersWithFallbacks(requestedParameters).filter(param => availableParameters.includes(param));
}

function getParametersWithFallbacks(parameterList)
{
    let completeList = [];
    for(let i=0;i<parameterList.length;i++)
    {
        completeList = completeList.concat(parameterList[i], fallbackParams[parameterList[i]]);
    }
    return completeList;
}

function arePragmaticWeatherParametersInList(requestedParameters, availableParameters)
{
    for(let i=0; i<requestedParameters.length;i++)
    {
        let requestedParameter = requestedParameters[i];
        // 1. Is the requested parameter in the availableParameters list?
        if(! availableParameters.includes(requestedParameter))
        {
            if(isEmpty(fallbackParams[requestedParameter]))
            {
                return false;
            }
            // If no, check if any fallback parameters are available
            let match = false;
            
            for(let j=0;j<fallbackParams[requestedParameter].length; j++)
            {
                if(availableParameters.includes(fallbackParams[requestedParameter][j]))
                {
                    match = true;
                }
            }
            if(!match)
            {
                return false;
            }
        }
    }

    return true;
}

function getRequiredParameters(modelMetaData)
{
    let parameterList = []
    modelMetaData.input.weather_parameters.forEach(function(weatherParameter){
        parameterList.push(weatherParameter.parameter_code)
    })
    return parameterList;
}

/**
 * 
 * @param {Array} coordinate [longitude,latitude] decimal degrees (WGS84)
 * @param {Json} datasource metadata
 * @returns 
 */
async function isPointCoveredByDatasource(coordinate, datasource)
{
    // If the geoJson is {"type":"Sphere"}, return true
    if(datasource.spatial.geoJSON != null && JSON.parse(datasource.spatial.geoJSON).type.toLowerCase() == "sphere")
    {
        return true;
    }
    let geoJson = await getDatasourceFeatures(JSON.parse(datasource.spatial.geoJSON), datasource.spatial.countries);
    let retVal = false;
    for(let i=0; i<geoJson.features.length;i++)
    {
        if(turf.booleanPointInPolygon(coordinate, geoJson.features[i]))
        {
            retVal = true;
        }
    }
    return retVal;
}

/**
 * Displays the weather data in a table
 * @param {Object} weatherData the weatherData in IPM Decisions format
 * @param {HTMLElement} container the element to render to
 * @param {String} tableClass for styling the table, e.g. using Bootstrap
 */ 
function renderWeatherData(weatherData, container, tableClass){
    
    let dates = getDateArray(weatherData.timeStart, weatherData.interval, weatherData.locationWeatherData[0].data.length);
    let html = "<table" + (tableClass != undefined ? " class=\"" + tableClass + "\"" : "") + "><thead><tr><th>Time</th>";
    weatherData.weatherParameters.forEach(function(weatherParameterId){
                    html +="<th>" + getWeatherParameter(weatherParameterId).name + "</th>";
                });
    
    html += "</tr></thead>"
    html += "<tbody>";
    for(let i=0;i< weatherData.locationWeatherData[0].data.length;i++)
    {
        html += "<tr><td>" + moment(dates[i]).format("YYYY-MM-DD HH:mm:ss") + "</td>";
            weatherData.locationWeatherData[0].data[i].forEach(function(val){html+="<td>" + val + "</td>";})
        html += "</tr>";
    }
    html += "</tbody>";
    html += "</table>";
    container.innerHTML = html;

}

/**
 * Merges the two datasets. Keeping the primaryData if both sets have data for same time and parameter
 * If a merge operation is not possible, primaryData is returned unchanged
 * @param {Object} primaryData 
 * @param {Object} secondaryData 
 */
function mergeWeatherData(primaryData, secondaryData)
{
    // Input check
    // If one set is null, return the non-null one unchanged
    // If both sets are null, return null
    if(primaryData == null || secondaryData == null)
    {
        return primaryData == null ? secondaryData : primaryData;
    }

    // If the log interval differs, return primaryData
    if(primaryData.interval != secondaryData.interval)
    {
        return primaryData;
    }

    // Initialize the result dataset
    let mergedData = {
        timeStart: moment(primaryData.timeStart <= secondaryData.timeStart ? primaryData.timeStart : secondaryData.timeStart),
        timeEnd: moment(primaryData.timeEnd >= secondaryData.timeEnd ? primaryData.timeEnd : secondaryData.timeEnd),
        interval: primaryData.interval,
        weatherParameters: function() {
            // Copy primaryData's weather parameters
            var mergedParams = [];
            for(let j=0;j<primaryData.weatherParameters.length;j++)
            {
                mergedParams.push(primaryData.weatherParameters[j]);
            };
            return mergedParams;
        }(),
        locationWeatherData: [
            {
                longitude: primaryData.locationWeatherData[0].longitude,
                latitude: primaryData.locationWeatherData[0].latitude
            }
        ]
    };

    // The parameter ordering may differ between primaryData and secondaryData
    // We solve this by mapping the parameter in secondaryData to an index in primaryData
    paramIndexes = {};
    secondaryData.weatherParameters.forEach(element => {
        if(mergedData.weatherParameters.includes(element))
        {
            paramIndexes[element] = mergedData.weatherParameters.indexOf(element);
        }
        // Checking for fallback parameters
        else if(Object.keys(fallbackParams).includes(element.toString()))
        {
            fallbackParams[element.toString()].forEach(fbElement=>{
                if(mergedData.weatherParameters.includes(fbElement))
                {
                    paramIndexes[element] = mergedData.weatherParameters.indexOf(fbElement);
                }
            });
        }
    });

    // Calculate dimensions of the merged data set
    let length = (getUnix(mergedData.timeEnd) - getUnix(mergedData.timeStart)) / mergedData.interval;
    let width = mergedData.weatherParameters.length;
    //console.info(length + "*" + width);
    let dataSet = Array.from(Array(length), () => new Array(width));
    // Keeping track of offsets (if dataset does not cover the entire combined period of the merging datasets)
    let primaryOffset = (getUnix(primaryData.timeStart) - getUnix(mergedData.timeStart)) / mergedData.interval;
    let secondaryOffset = (getUnix(secondaryData.timeStart) - getUnix(mergedData.timeStart)) / mergedData.interval;
    let primaryDataSet = primaryData.locationWeatherData[0].data;
    let secondaryDataSet = secondaryData.locationWeatherData[0].data;

    // Finally: Merge the data!
    for(let i=0;i<dataSet.length;i++)
    {

        if(secondaryOffset <= i && i-secondaryOffset < secondaryDataSet.length)
        {
            for(let j=0; j<secondaryData.weatherParameters.length;j++)
            {
                if(paramIndexes[secondaryData.weatherParameters[j]] != undefined && !isEmpty(secondaryDataSet[i-secondaryOffset][j]))
                {
                    dataSet[i][paramIndexes[secondaryData.weatherParameters[j]]] = secondaryDataSet[i-secondaryOffset][j];
                }
            }
        }
        if(primaryOffset <= i && i-primaryOffset < primaryDataSet.length )
        {
            for(let j=0; j<primaryData.weatherParameters.length;j++)
            {
                if(!isEmpty(primaryDataSet[i-primaryOffset][j]))
                {
                    dataSet[i][j] = primaryDataSet[i-primaryOffset][j];
                }
            }
        }
    }

    mergedData.locationWeatherData[0].data = dataSet;
    trimWeatherData(mergedData);
    return mergedData;
}

/**
 * Removes NULL rows at the beginning and end of a dataset
 * Assumes only one locationWeatherData
 * @param {Object} weatherData 
 */
function trimWeatherData(weatherData)
{
    // Find index of first and last non-NULL rows
    let firstIndex = 0;
    outer_loop_first:
    for(let i=0; i<weatherData.locationWeatherData[0].data.length; i++)
    {
        let row = weatherData.locationWeatherData[0].data[i];
        if(row != undefined && row != null )
        {
            for(let j=0;j<row.length;j++)
            {
                if(row[j] != undefined && row[j] != null)
                {
                    firstIndex = i;
                    break outer_loop_first;
                }
            }
        }
    }

    let lastIndex = weatherData.locationWeatherData[0].data.length -1;
    outer_loop_last:
    for(let i=lastIndex; i>=0; i--)
    {
        let row = weatherData.locationWeatherData[0].data[i];
        if(row != undefined && row != null )
        {
            for(let j=0;j<row.length;j++)
            {
                if(row[j] != undefined && row[j] != null)
                {
                    lastIndex = i;
                    break outer_loop_last;
                }
            }
        }
    }

    

    // Adjust timeStart and timeEnd based on indexes
    let adjustedTimeStart = moment(weatherData.timeStart).add(firstIndex, (weatherData.interval == 86400 ? 'days' : 'hours'));
    let adjustedTimeEnd   = moment(weatherData.timeEnd).subtract(weatherData.locationWeatherData[0].data.length -1 - lastIndex, (weatherData.interval == 86400 ? 'days' : 'hours'));
    
    weatherData.timeStart = adjustedTimeStart;
    weatherData.timeEnd = adjustedTimeEnd;
    weatherData.locationWeatherData[0].data = weatherData.locationWeatherData[0].data.slice(firstIndex, lastIndex + 1);
}

/**
 * Util method if you don't know the type of object you're trying to get a Unix timestamp from
 * @param {Object} aDate 
 * @returns {BigInt} the Unix timestamp
 */
function getUnix(aDate)
{
    if(typeof aDate == "string")
    {
        aDate = moment(aDate);
    }
    return aDate.unix();
}

/**
 * Saves a bit of typing. Currently checks if the val is either undefined or null, and if a string == ""
 * @param {Object} val 
 * @returns {Boolean} true if the value is considered empty
 */
function isEmpty(val)
{
    return val === undefined || val === null || (typeof val === "string" && val.trim() == "");
}

// Keeping track of maps
var maps = {};

/**
 * 
 * @param {String} containerId the container where the map to destroy is anchored
 */
function destroyDataSourceMap(containerId)
{
    let map = maps[containerId];
    // Delete whatever was there
    if(map != undefined)
    {
        map.setTarget(undefined);
        map = null;
    }
}

/**
 * Requires OpenLayers v. 4
 * @param {String} containerId
 * @param {JSON} geoJson 
 * @param {Array} countryCodeList
 * @param {Boolean} clickable if null or false; clicking does nothing
 * @param {Function} featureClickedCallback Callback function taking parameters ({String} id, {Array<Float>} coordinate). If null => do nothing when user clicks in map
 */
async function initDataSourceMap(containerId, geoJson, countryCodeList, featureClickedCallback)
{
    
    let map = maps[containerId];
    // Delete whatever was there
    if(map != undefined)
    {
        map.setTarget(undefined);
        map = null;
    }

    // If no data, no map
    if((isEmpty(geoJson) || ((isEmpty(geoJson.features) || geoJson.features.length == 0) && ! (geoJson.type != undefined && geoJson.type.toLowerCase() == "sphere")) ) && (countryCodeList == null || countryCodeList.length == 0))
    {
        document.getElementById(containerId).innerHTML = "NO GEODATA PROVIDED BY WEATHER DATA SOURCE";
        return;
    }
    else
    {
        document.getElementById(containerId).innerHTML = "";
    }

    // Using OpenStreetMap as default layer
    var backgroundLayer = new ol.layer.Tile({
        source: new ol.source.OSM()
    });

    // Creating the map
    map = new ol.Map({
            target: containerId,
            layers: [backgroundLayer],
            renderer: 'canvas'
    });


    // Adding it to list of maps, for bookkeeping
    maps[containerId] = map;

    // Setting zoom and center for the map (need to do this after creating map. so that we can transform our
	// center to correct map projection)
	var view = new ol.View({
		center: ol.proj.transform([10,65], 'EPSG:4326', map.getView().getProjection().getCode()),
		zoom: 7
	});
	map.setView(view);
	
    // Read the datasource's geoJson features
    let drawnFeatures = undefined;
    let features = new ol.Collection();
    let format = new ol.format.GeoJSON();
    drawnFeatures = await format.readFeatures(
        await getDatasourceFeatures(geoJson, countryCodeList), 
        {
            dataProjection: 'EPSG:4326',
            featureProjection: map.getView().getProjection().getCode()
        }
    );

    // Add any features to the featureOverlay
    let featureOverlay = undefined;
    
    // Create an empty layer
    featureOverlay = new ol.layer.Vector({
        source: new ol.source.Vector({
        features: features
        }),
        style: styleUnselected
    });
    // Add the stations or area
    if(drawnFeatures != undefined && drawnFeatures.length > 0)
    {
        featureOverlay.getSource().addFeatures(drawnFeatures);
        // Fit the features to the extent of the map
        let extent = featureOverlay.getSource().getExtent();
        map.getView().fit(extent, map.getSize());
    }
    // If no stations / area: show the world
    else
    {
        map.getView().setZoom(1);
        map.getView().setCenter(ol.proj.transform([0.0,15.0], 'EPSG:4326', map.getView().getProjection().getCode()));
        
    }
    featureOverlay.setMap(map);    
    
    
 
    if(featureClickedCallback != undefined && featureClickedCallback != null)
    {
        /*
        * A little bit of this and that to make feature selection work the way we want
        * - Highlight a selected station (but only one station, multiple select is not allowed)
        * - If user clicks on multiple stations, zoom in to make it easy to pick the right one
        */ 

        let selectInteraction = new ol.interaction.Select({
            toggleCondition: ol.events.condition.never // Only one can be selected at a time
        });
        map.addInteraction(selectInteraction);
        selectInteraction.on("select", function(e){
            // If it's an area (id == undefined), don't mark it as selected
            if(e.target.getFeatures().getLength() > 0 && e.target.getFeatures().item(0).getId() == undefined)
            {
                e.target.getFeatures().clear();
            }
            else if(e.target.getFeatures().getLength() == 1)
            {
                let feature = e.target.getFeatures().item(0);
                featureClickedCallback(feature.getId(), undefined);
            }
        });
    
        // We add this in order to detect if multiple features were clicked on, since the selectInteraction
        // does not catch that
        map.on('singleclick', function(evt) {
            var pixel = map.getEventPixel(evt.originalEvent);
            var coordinate = map.getEventCoordinate(evt.originalEvent);
            handleMapClicked(map, featureOverlay, pixel, coordinate, featureClickedCallback);
        });
    }
}

/**
 * Get "complete" spatial info from weather datasource, using either provided GeoJson or 
 * inferred from list of countries
 * @param {JSON} geoJson 
 * @param {Array} countryCodeList 
 * @returns {Json} GeoJson
 */
async function getDatasourceFeatures(geoJson, countryCodeList)
{
    if(geoJson != null && geoJson.type != undefined && geoJson.type.toLowerCase() == "sphere")
    {
        return {"type": "FeatureCollection","features": []};
    }
    // If we have geoJson available, we display that
    if (geoJson != null  && geoJson.features !== undefined && geoJson.features.length > 0) {
        return geoJson;
    }
    return await getCountryBoundaries(countryCodeList);
}

// Default style for weather stations and areas
const styleUnselected = new ol.style.Style({
    fill: new ol.style.Fill({
        color: 'rgba(173, 173, 173, 0.5)'
      }),
      stroke: new ol.style.Stroke({
        color: '#000000',
        width: 1
      }),
      image: new ol.style.Circle({
        radius: 6,
        fill: new ol.style.Fill({
          color: '#adadad'
        }),
        stroke: new ol.style.Stroke({
            color: '#000000',
            width: 1
          })
      })
});

/*
 * Singleton pattern works. Used in function below
 */



/**
 * When a user clicks on the map, handles whatever was clicked (station, point in area, point outside any feature/area)
 * @param {ol.map} map 
 * @param {ol.layer} layer
 * @param {ol.Pixel} pixel 
 * @param {ol.coordinate} coordinate 
 * @param {Function} featureClickedCallback Callback function taking parameters ({String} id, {Array<Float>} coordinate)
 */
function handleMapClicked(map, layer, pixel, coordinate, featureClickedCallback) {
	var features = [];
    map.forEachFeatureAtPixel(pixel, function(feature){
       features.push(feature); 
    });

    if (features.length == 1 || layer.getSource().getFeatures().length == 0) {
        // Area / sphere (whole globe)
        if(layer.getSource().getFeatures().length == 0 || features[0].getId() === undefined)
        {
            // We must place a marker where the user clicked
            let marker = map.getOverlayById("marker");
            // If first time click, create a new marker
            if(marker == undefined)
            {
                map.addOverlay(new ol.Overlay({
                        id: "marker",
                        element: function() {
                            let markerElement = document.createElement("div");
                            markerElement.setAttribute(
                                "style",
                                "width: 20px; height: 20px; border: 1px solid #088; border-radius: 10px; background-color: #0FF; opacity: 0.5;"
                            );
                            return markerElement
                        }(),
                        position: coordinate,
                        positioning: "center-center"
                    })
                );
            }
            // Otherwise, just move the current marker
            else
            {
                marker.setPosition(coordinate);
            }
            featureClickedCallback(null, getDecimalDegrees(map, coordinate));
        }	
        // Click on a station is handled in the select interaction (see initDataSourceMap)
    }
    // The select interaction (see initDataSourceMap) highlights one and only one
    // selected feature. If the clicked pixel has more than one feature, 
    // zoom in to let the user be able to separate the features
    else if(features.length > 1)
    {
        map.getView().setCenter(coordinate);
        map.getView().setZoom(map.getView().getZoom() + 2);
    }
};

function createMarkerOverlay(coordinate)
{
    
}

/**
 * 
 * @param {ol.Map} map 
 * @param {Array<Float>} coordinate 
 * @return {ol.Coordinate} the pixel coordinate from a map converted to WGS84 Decimal degrees
 */
function getDecimalDegrees(map, coordinate)
{
    return ol.proj.transform(coordinate, map.getView().getProjection().getCode(), 'EPSG:4326');
}

async function getCountryBoundaries(countryCodeList)
{
    if(countryCodeList == undefined || countryCodeList == null)
    {
        return {};
    }
    const response = await fetch(ipmdWeatherApiURL + "rest/country/" + countryCodeList.join(","));
    return await response.json();
}


/**
 * List of fallback parameters to use if the weather data source does not provide the 
 * initially requested weather parameter
 */
const fallbackParams = {
        1001: [1002],
        1002: [1001],
        3001: [3002],
        3002: [3001],
        4002: [4003,4012,4013],
        4003: [4002,4013,4012],
        4012: [4013,4002,4003],
        4013: [4012,4003,4002]
}

/**
 * Util method for programmatically selecting an option in a list
 * @param {HTMLElement} selectList 
 * @param {String} optionValue 
 */
function setSelection(selectList, optionValue)
{
    for(let i=0;i<selectList.options.length; i++)
    {
        if(selectList.options[i].value == optionValue)
        {
            selectList.options[i].selected = true;
        }
        else
        {
            selectList.options[i].selected = false;
        }
    }
}


/**
 * All the available weather parameters
 */
const weatherParameterList = [
    {
        "id": 1001,
        "name": "Instantaneous temperature at 2m",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1002,
        "name": "Mean air temperature at 2m",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1003,
        "name": "Minimum air temperature at 2m",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "MIN"
    },
    {
        "id": 1004,
        "name": "Maximum air temperature at 2m",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "MAX"
    },
    {
        "id": 1021,
        "name": "Instantaneous temperature in canopy",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1022,
        "name": "Mean air temperature in canopy",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1023,
        "name": "Minimum air temperature in canopy",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "MIN"
    },
    {
        "id": 1024,
        "name": "Maximum air temperature in canopy",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "MAX"
    },
    {
        "id": 1101,
        "name": "Instantaneous temperature at -5cm (Celcius)",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1102,
        "name": "Mean temperature at -5cm",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1111,
        "name": "Instantaneous temperature at -10cm (Celcius)",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1112,
        "name": "Mean temperature at -10cm",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1121,
        "name": "Instantaneous temperature at -20cm (Celcius)",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1122,
        "name": "Mean temperature at -20cm",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1131,
        "name": "Instantaneous temperature at -30cm (Celcius)",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1132,
        "name": "Mean temperature at -30cm",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1141,
        "name": "Instantaneous temperature at -40cm (Celcius)",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1142,
        "name": "Mean temperature at -40cm",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1151,
        "name": "Instantaneous temperature at -50cm (Celcius)",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1152,
        "name": "Mean temperature at -50cm",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 1901,
        "name": "Dew point temperature",
        "description": null,
        "unit": "Celcius",
        "aggregationType": "AVG"
    },
    {
        "id": 2001,
        "name": "Precipitation",
        "description": null,
        "unit": "mm",
        "aggregationType": "SUM"
    },
    {
        "id": 3001,
        "name": "Instantaneous RH at 2m (%)",
        "description": null,
        "unit": "%",
        "aggregationType": "AVG"
    },
    {
        "id": 3002,
        "name": "Mean RH at 2m",
        "description": null,
        "unit": "%",
        "aggregationType": "AVG"
    },
    {
        "id": 3003,
        "name": "Minimum RH at 2m",
        "description": null,
        "unit": "%",
        "aggregationType": "MIN"
    },
    {
        "id": 3004,
        "name": "Maximum RH at 2m",
        "description": null,
        "unit": "%",
        "aggregationType": "MAX"
    },
    {
        "id": 3021,
        "name": "Instantaneous RH in canopy",
        "description": null,
        "unit": "%",
        "aggregationType": "AVG"
    },
    {
        "id": 3022,
        "name": "Mean RH in canopy",
        "description": null,
        "unit": "%",
        "aggregationType": "AVG"
    },
    {
        "id": 3023,
        "name": "Minimum RH in canopy",
        "description": null,
        "unit": "%",
        "aggregationType": "MIN"
    },
    {
        "id": 3024,
        "name": "Maximum RH in canopy",
        "description": null,
        "unit": "%",
        "aggregationType": "MAX"
    },
    {
        "id": 3101,
        "name": "Leaf wetness in 2m (minutes/hour)",
        "description": null,
        "unit": "minutes/hour",
        "aggregationType": "SUM"
    },
    {
        "id": 3102,
        "name": "Leaf wetness in canopy",
        "description": null,
        "unit": "minutes/hour",
        "aggregationType": "SUM"
    },
    {
        "id": 3103,
        "name": "Leaf wetness in grass",
        "description": null,
        "unit": "minutes/hour",
        "aggregationType": "SUM"
    },
    {
        "id": 4001,
        "name": "Wind direction at 2m (degrees 0-360)",
        "description": null,
        "unit": "degrees",
        "aggregationType": "AVG"
    },
    {
        "id": 4002,
        "name": "Instantaneous wind speed at 2m",
        "description": null,
        "unit": "m/s",
        "aggregationType": "AVG"
    },
    {
        "id": 4003,
        "name": "Mean wind speed at 2m",
        "description": null,
        "unit": "m/s",
        "aggregationType": "AVG"
    },
    {
        "id": 4004,
        "name": "Max wind speed at 2m",
        "description": null,
        "unit": "m/s",
        "aggregationType": "MAX"
    },
    {
        "id": 4005,
        "name": "Min wind speed at 2m",
        "description": null,
        "unit": "m/s",
        "aggregationType": "MIN"
    },
    {
        "id": 4011,
        "name": "Wind direction at 10m (degrees 0-360)",
        "description": null,
        "unit": "degrees",
        "aggregationType": "AVG"
    },
    {
        "id": 4012,
        "name": "Instantaneous wind speed at 10m",
        "description": null,
        "unit": "m/s",
        "aggregationType": "AVG"
    },
    {
        "id": 4013,
        "name": "Mean wind speed at 10m",
        "description": null,
        "unit": "m/s",
        "aggregationType": "AVG"
    },
    {
        "id": 4014,
        "name": "Max wind speed at 10m",
        "description": null,
        "unit": "m/s",
        "aggregationType": "MAX"
    },
    {
        "id": 4015,
        "name": "Min wind speed at 10m",
        "description": null,
        "unit": "m/s",
        "aggregationType": "MIN"
    },
    {
        "id": 5001,
        "name": "Solar radiation (Q0) (W/sqm)",
        "description": null,
        "unit": "W/sqm",
        "aggregationType": "SUM"
    }
];